/*
  The partysip program is a modular SIP proxy server (SIP -rfc3261-)
  Copyright (C) 2002,2003  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002,2003  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _PSP_REQ_H
#define _PSP_REQ_H

#ifndef WIN32
#include "config.h"
#endif

#include <ppl/ppl_time.h>
#include <ppl/ppl_dso.h>
#include <ppl/ppl_dns.h>

#include <osip2/osip.h>
#include <osipparser2/osip_parser.h>
#include <osip2/osip_mt.h>

#include <partysip/psp_macros.h>

/* psp_req_t is
   1: created by the imp module upon reception of a new request
   2: given to the plugins sequentially
   3: updated by plugins
   4: given in the fifo of ups, slr or sfp
   Final Information is used to answer or forward the transaction
*/


typedef struct sfp_branch_t sfp_branch_t;

struct sfp_branch_t
{
  char *branch;
  ppl_time_t ctx_start;
  int timer_c;
  int ctx_timeout;
  osip_transaction_t *out_tr;
  osip_uri_t *url;
  ppl_dns_ip_t *fallbacks;
  osip_message_t *request;
  osip_message_t *response;
  int last_code_forwarded;
  sfp_branch_t *next;
  sfp_branch_t *parent;
};

int sfp_branch_init (sfp_branch_t ** sfp_branch, char *branch);
void sfp_branch_free (sfp_branch_t * branch);
int sfp_branch_get_osip_id (sfp_branch_t * branch);
int sfp_branch_set_transaction (sfp_branch_t * branch,
				osip_transaction_t * out_tr);
int sfp_branch_set_timeout (sfp_branch_t * branch, int delay);
int sfp_branch_is_expired (sfp_branch_t * branch);
int sfp_branch_set_url (sfp_branch_t * branch, osip_uri_t * url);
int sfp_branch_set_request (sfp_branch_t * branch, osip_message_t * req);
int sfp_branch_set_response (sfp_branch_t * branch, osip_message_t * resp);
int sfp_branch_get_last_code_forwarded (sfp_branch_t * sfp);
int sfp_branch_set_last_code_forwarded (sfp_branch_t * sfp, int lastcode);
int sfp_branch_clone_and_set_fallbacks (sfp_branch_t * branch,
					ppl_dns_ip_t * ips);

typedef struct _rsp_ctx_t
{
  int fork;			/* 1: use fork capabilities, 0: use sequential search */
  int ind;			/* next via number for outgoing branch */
  char *branch;
  sfp_branch_t *branches;

  int state;
#define RSPCTX_CALLING 0
  /* psp_req is just arrived and no branch exist.
     action1: When new locations are resolved, the transaction is initiated.
     action2: Look in the list of cancel requests, find the related transaction
     and send a 407 To INVITE and a 200 to CANCEL, go in the
     RSP_CTX_CANCELLED STATE.
   */
#define RSPCTX_PROCEEDING 1
  /* some branches exist.
     action1: When new locations are resolved, the transaction is initiated. 
     action2: When an outgoing transaction is in the ICT_COMPLETED or
     NICT_COMPLETED state, we forward the final response and go in the CANCELLING
     state.
     action2: When an outgoing transaction is in the ICT_PROCEEDING or
     NICT_PROCEEDING state, we forward the provisionnal response and
     remain in the started state. (The message must only be forwarded
     once!!!)
     action3: Look in the list of cancel requests, find the related transaction
     and send a 407 To INVITE and a 200 to CANCEL, go in the
     RSP_CTX_CANCELLED STATE.
   */
#define RSPCTX_COMPLETED 3
#define RSPCTX_TERMINATED 4
  /* at least one branch is answered, the others must be CANCELLED...
     actions1: For all transactions in NICT_PROCEEDING or ICT_PROCEEDING
     send a CANCEL transaction.
   */
#define RSPCTX_CANCELLED 5
  /*
     action1: ,
     and send a cancel on all branch in the ICT_PROCEEDING transaction. No more
     to do...
   */
#define RSPCTX_BROKEN 6
  /* FOR ALL STATE!!
     action1:
     wait for all transactions to timeout (in the *T_TERMINATED state or birth_time
     is more than 4 minutes for INVITEs and 60s for other REQUESTs...)
   */

}
rsp_ctx_t;

int rsp_ctx_init (rsp_ctx_t ** rsp_ctx, char *branch);
void rsp_ctx_free (rsp_ctx_t * branch);
int rsp_ctx_set_fork (rsp_ctx_t * branch);
int rsp_ctx_set_sequential (rsp_ctx_t * rsp_ctx);
int rsp_ctx_set_state (rsp_ctx_t * rsp_ctx, int state);
int rsp_ctx_get_state (rsp_ctx_t * rsp_ctx);

typedef struct location_t location_t;

struct location_t
{
  osip_uri_t *url;			/* THIS SHOULD BE CHANGED TO A CONTACT HEADER!!! */
  char  *path;			/* add a path information (see rfc3327.txt) */
  int expire;
  location_t *next;
  location_t *parent;
};

PPL_DECLARE (int)
location_init (location_t ** location, osip_uri_t * url, int expire);
PPL_DECLARE (void)
location_free (location_t * location);
PPL_DECLARE (int)
location_set_url (location_t * location, osip_uri_t * url);
PPL_DECLARE (int)
location_set_expires (location_t * location, int expire);
PPL_DECLARE (int)
location_set_path (location_t * location, char *path);
PPL_DECLARE (int)
location_clone (location_t * loc, location_t ** loc2);


/* extra stuff for dialog between core layer and plugins.
   plugin id: unique number for plugin.
   request id: transaction id(?) that triggered this request.
*/
#define CTX_READY           0x01
#define CTX_NOT_READY       0x02
#define CTX_PARTIAL         0x04	/* in later step? */

/*
typedef struct plug_ctx_t plug_ctx_t;

struct plug_ctx_t {
  int   plug_id;              this context actually refer to a plugin

  int   plug_flg;            NOT_READY | READY
  void *plug_ctx;            your plugin context that will be sent back

  location_t *locations;     for plugin who has search location capabilities

  plug_ctx_t *next;
  plug_ctx_t *parent;
};

int  plug_ctx_init(plug_ctx_t **ctx, int plug_id, int flag);
void plug_ctx_free(plug_ctx_t *ctx);
int  plug_ctx_add_location(plug_ctx_t *ctx, location_t *loc);
 */

     typedef struct psp_req_t psp_req_t;

     struct psp_req_t
     {
       /*  int osip_id;     obsolete: this id is the one of the incoming transaction. */
       osip_transaction_t *inc_tr;	/* incoming transaction */
       osip_message_t *request;		/* initial request trigerring the event   */
       osip_message_t *response;		/* response trigerring the event          */
       time_t start;		/* creation date of element               */
#ifndef TIMEOUT_FOR_NO_ANSWER
#define TIMEOUT_FOR_NO_ANSWER 15
#endif
#ifndef TIMEOUT_FOR_INVITE
#define TIMEOUT_FOR_INVITE 120
#endif
#ifndef TIMEOUT_FOR_NONINVITE
#define TIMEOUT_FOR_NONINVITE 32
#endif
       int timeout;
       int owners;		/* number of owners of this context       */
       /* if plugin_owners=0 : this context is not linked to any plugins */

       /* mode flag */
#define PSP_UAS_MODE        0x001
#define PSP_SFULL_MODE      0x008

       /* state flag */
#define PSP_CONTINUE        0x010
#define PSP_PROPOSE         0x020
#define PSP_MANDATE         0x040
#define PSP_STOP            0x080

       /* mode properties */
#define PSP_STAY_ON_PATH    0x100

#define __IS_PSP_UAS_MODE(flag)   ((~flag | ~PSP_UAS_MODE ) == ~PSP_UAS_MODE)
#define __IS_PSP_SFULL_MODE(flag) ((~flag | ~PSP_SFULL_MODE ) == ~PSP_SFULL_MODE)

#define __SET_UAS_MODE(flag)      (flag = ((flag&0xFF0)|PSP_UAS_MODE))
#define __SET_SFULL_MODE(flag)    (flag = ((flag&0xFF0)|PSP_SFULL_MODE))

#define __IS_PSP_CONTINUE(flag)   ((~flag | ~PSP_CONTINUE ) == ~PSP_CONTINUE)
#define __IS_PSP_PROPOSE(flag)    ((~flag | ~PSP_PROPOSE  ) == ~PSP_PROPOSE)
#define __IS_PSP_MANDATE(flag)    ((~flag | ~PSP_MANDATE  ) == ~PSP_MANDATE)
#define __IS_PSP_STOP(flag)       ((~flag | ~PSP_STOP  ) == ~PSP_STOP)

#define __SET_CONTINUE(flag)      (flag = ((flag&0xF0F)|PSP_CONTINUE))
#define __SET_PROPOSE(flag)       (flag = ((flag&0xF0F)|PSP_PROPOSE))
#define __SET_MANDATE(flag)       (flag = ((flag&0xF0F)|PSP_MANDATE))
#define __SET_STOP(flag)          (flag = ((flag&0xF0F)|PSP_STOP))

#define __SET_STAY_ON_PATH(flag)    (flag = ((flag&0x0FF)|PSP_STAY_ON_PATH))
#define __IS_PSP_STAY_ON_PATH(flag) ((~flag|~PSP_STAY_ON_PATH)==~PSP_STAY_ON_PATH)

       int flag;		/* contains state and mode flags */
       int uas_status;		/* UAS return code proposed or mandated */

       location_t *locations;
       /*  plug_ctx_t   *plug_ctxs;  IS IT REALLY USEFULL? we'll see if some plugins
          can't do without it. */


       rsp_ctx_t *rsp_ctx;	/* used by statefull module */

       psp_req_t *next;
       psp_req_t *parent;
     };

     int psp_req_init (psp_req_t ** req, osip_transaction_t * inc_tr);
     void psp_req_free (psp_req_t * req);
PPL_DECLARE (osip_transaction_t *) psp_req_get_incoming_transaction (psp_req_t *
								req);
PPL_DECLARE (int)
psp_req_set_state (psp_req_t * req, int state);
PPL_DECLARE (int)
psp_req_set_mode (psp_req_t * req, int mode);
PPL_DECLARE (int)
psp_req_set_property (psp_req_t * req, int property);
PPL_DECLARE (int)
psp_req_set_uas_status (psp_req_t * req, int status);

/* int  psp_req_add_plugin_ctx(psp_req_t *req, plug_ctx_t *ctx); */
     int psp_req_set_response_context (psp_req_t * req, rsp_ctx_t * ctx);
     int psp_req_take_ownership (psp_req_t * req);
     int psp_req_release_ownership (psp_req_t * req);
PPL_DECLARE (int)
psp_req_get_timeout (psp_req_t * req);
PPL_DECLARE (int)
psp_req_add_timeout (psp_req_t * req, int add_sec);
PPL_DECLARE (int)
psp_req_set_timeout (psp_req_t * req, int timeout);
PPL_DECLARE (int)
psp_req_is_expired (psp_req_t * req);
     sfp_branch_t *psp_core_sfp_find_branch (psp_req_t * req, int osip_id);

PPL_DECLARE (int)
psp_req_get_uas_status (psp_req_t * req);

/* continue the accessor API */

#endif
